#! /usr/bin/env python
#
#   Copyright 2011 S. Brewster Malevich <malevich@email.arizona.edu>
#
#   pydendro is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>

"""
Classes to read dendrochronological files from the ITRDB.
"""

import os
import linecache
from scipy.stats.stats import pearsonr
import numpy as np

class DendroFile:
    """Parent class to be inherited specific dendro files formats.
    
    Arguments:
    path -- Path to the target .crn file.

    """
    def __init__(self, path):
        """Opens file 'path' and checks that it exists.
        """
        if not os.path.isfile(path):
            raise IOError()
        self._path = path

class Tucson(DendroFile):
    """Class to handle .rwl files from the ITRDB.
    
    -This is incomplete-

    Arguments:
    path -- Path to the target .crn file.

    """
    def __init__(self, path):
        """Parses the RWL file, defining instance attributes.
        """
        DendroFile.__init__(self,path)


class Chronology(DendroFile):
    """Class to handle *.crn files from the ITRDB.
    
    Arguments:
    path -- Path to the target .crn file.
    
    Returns:
    An object of class Chronology.
    
    
    Note that the optional ITRDB "Chronology Statistics" will not be
    preserved. This object is based on format definitions of the ITRDB (see
    http://www.ncdc.noaa.gov/paleo/treeinfo.html).
   

    Instance attributes:

    type
        A string indicating the tree-ring chronology file type (e.g. "ARSTND",
        "Standard", etc.).

    years
        A list of integers corresponding to each year with a chronology index
        value.

    values
        A list of float for chronology tree-ring index values for each
        of the chronology's years.

    sampledepth
        An integer list of the measurement sample site for each year of the
        chronology.

    speciesname
        A string name of sampled species for the chronology site.

    speciescode
        A string code for the sampled species of the chronology site.

    siteid
        A string site ID for the chronology.

    sitename
        Name of the chronology's sample site. A string.

    location
        String giving the state and country of the chronology's sample site.

    latlon
        An integer of the chronology sample site's latitude and longitude 
        (ddmm or dddmm). I've found a variety of formats for this so, beware.

    elevation
        Integer of the chronology site's elevation.

    firstyear
        The first year of the chronology. An integer.

    lastyear
        The last year of the chronology. An integer.

    collectiondate
        A string giving the chonology's final collection date.

    investigator
        A string of lead investigator(s).

    """
    def __init__(self, path):
        """Parses the Chronology file defining instance attributes.
        """
        DendroFile.__init__(self,path)
        self.type = None
        self.years = []
        self.values = []
        self.sampledepth = []
        self.speciesname = None
        self.speciescode = None
        self.siteid = None
        self.sitename = None
        self.location = None
        self.latlon = None
        self.elevation = None
        self.firstyear = None
        self.lastyear = None
        self.collectiondate = None
        self.investigator = None
        victim = None
        types = {'a.crn':'ARSTND',
                 'p.crn':'Low Pass Filter',
                 'r.crn':'Residual',
                 's.crn':'Standard',
                 'w.crn':'Re-Whitened Residual',
                 'n.crn':'Measurements Only'}
        if self._path[-5:] in types:
            self.type = types[self._path[-5:]]
        else:
            self.type = 'Unknown'
        headermark = [' 1 ', ' 2 ', ' 3 ']  # To identify the header.
        header1 = linecache.getline(path, 1)
        header2 = linecache.getline(path, 2)
        header3 = linecache.getline(path, 3)
        if header1[6:9] == headermark[0]:
            self.siteid = str(header1[0:6]).strip()
            self.sitename = str(header1[9:61]).strip()
            self.speciescode = str(header1[61:65]).strip()
        else:
            raise NameError('File ' + self._path
                            + ' appears to have no header.')
        if header2[6:9] == headermark[1]:
            self.location = str(header2[9:22]).strip()
            self.speciesname = str(header2[22:30]).strip()
            # Slice elevation to ':44' to cut out the "M".
            self.elevation = int(header2[40:44])
            self.latlon = str(header2[47:57]).strip()
            self.firstyear = int(header2[67:71])
            self.lastyear = int(header2[72:76])
            self.years = range(self.firstyear, self.lastyear + 1)
        else:
            raise NameError('File ' + self._path 
                            + ' appears to have no header.')
        if header3[6:9] == headermark[2]:
            self.investigator = str(header3[9:72]).strip()
            self.collectiondate = str(header3[72:80]).strip()
        else:
            raise NameError('File ' + self._path
                            + ' appears to have no header.')
        lastdecade = self.lastyear - (self.lastyear % 10)  # To ID last line.
        for line in open(path):
            if line[6:9] in headermark:
                continue
            a = 10  # Tree-ring index index.
            b = 14
            c = 14  # Ring count index.
            d = 17
            for i in range(0, 10):
                victim = line[a:b]
                if victim == '9990':
                    a +=  7
                    b +=  7
                    c +=  7
                    d +=  7
                    continue
                victim = victim.strip()
                victim = victim.zfill(4)
                self.values.append(float(victim[:1] + '.' + victim[1:]))
                self.sampledepth.append(int(line[c:d]))
                a +=  7
                b +=  7
                c +=  7
                d +=  7
            if int(line[6:10]) == lastdecade:  # Stop after last line.
                break
        if len(self.values) != len(self.sampledepth) != len(self.years):
            raise NameError('Unequal values, samples sizes'
                            + 'and years from chronology file.')
    
    def crop_corr(self, x, x_years):
        """Crop and correlate a chrnology with a given variable.
        
        Arguments:
        x -- An array/list(?) of your variable of interest.
        x_years -- Array/list of years corresponding to the x observations.
    
        Returns:
        A tuple giving 1) Pearson's correlation coefficient, 2) 2-tailed
        p-value, 3) length of the common period between crn and x.
    
        """
        crn_values = np.array(self.values)
        crn_years = np.array(self.years)
        if len(crn_values) != len(crn_years):
            raise ValueError('unequal values and years')
        # Might want to check if x and x_years are not array...?
        # Type checks are python taboo but I'm not sure how numpy will handle
        # lists.
        year_min = np.maximum(np.min(crn_years), np.min(x_years))
        year_max = np.minimum(np.max(crn_years), np.max(x_years))
        crn_crop = crn_values[(crn_years >= year_min)
            & (crn_years <= year_max)]
        x_crop = x[(x_years >= year_min) & (x_years <= year_max)]
        results = pearsonr(crn_crop, x_crop)
        results += (len(crn_crop),)
        return results

